import logging

import numpy as np
from astropy.table import Table
from qtpy import QtWidgets, QtCore

from .AbstractWidget import AbstractWidget
from .table_tools import column_not_found_message


logger = logging.getLogger(__name__)


class SmartSlider(QtWidgets.QSlider):
    value_changed = QtCore.Signal()

    def __init__(self, min_value=0, max_value=100, step=1, default_value=0, **kwargs):

        super().__init__(**kwargs)

        self._n = int((max_value - min_value) / step) + 1
        self._arr = np.linspace(min_value, max_value, self._n)

        self.step = step
        self.default_value = default_value

        self.setRange(1, self._n)
        self.setSingleStep(1)
        self.setValue(self.default_index)

        self._index = self.default_index

        self.valueChanged[int].connect(self.value_changed_action)

    @property
    def default_index(self):
        try:
            return self.index_from_value(self.default_value)
        except (TypeError, ValueError):
            return 1

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, i):
        self._index = i
        self.setValue(i)

    @property
    def value(self):
        return self._arr[self.index - 1]

    def index_from_value(self, value):
        i = self._arr.searchsorted(value, side='right')
        return i if i > 0 else 1

    def reset(self):
        self.index = self.default_index

    def value_changed_action(self, index: int):
        self.index = index
        self.value_changed.emit()


class SmartSliderWithEditor(AbstractWidget):
    value_changed = QtCore.Signal(float)

    def __init__(self, parameter: str = 'x', full_name: str | None = None, action: str | None = None,
                 visible: bool = True, catalogue_name: str | None = None, show_text_editor: bool = False,
                 text_editor_precision: int = 6, parent=None, **kwargs):

        super().__init__(layout=QtWidgets.QGridLayout(), parent=parent)

        self.layout().setSpacing(5)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.setHidden(not visible)

        self.parameter = parameter if catalogue_name is None else catalogue_name
        self.full_name = parameter if full_name is None else full_name
        self.action = f"change {self.full_name}" if action is None else action

        self.cat_name = catalogue_name
        self.text_editor = show_text_editor
        self.precision = text_editor_precision

        # create a slider
        self._slider = SmartSlider(orientation=QtCore.Qt.Orientation.Horizontal if self.text_editor
                                   else QtCore.Qt.Orientation.Vertical, parent=self, **kwargs)
        self._slider.value_changed.connect(self.update_from_slider)
        self._slider.setToolTip(f'Slide to {self.action}')

        self._default_value_backup = self._slider.default_value

        # create a label
        self._label = QtWidgets.QLabel(f'{self.parameter} =', parent=self)

        # create a text editor
        self._editor = QtWidgets.QLineEdit(parent=self)
        self._editor.returnPressed.connect(self._update_from_editor)
        self._editor.setMaximumWidth(120)

        self._label.setHidden(not show_text_editor)
        self._editor.setHidden(not show_text_editor)

    def init_ui(self):
        self.layout().addWidget(self._slider, 1, 1, 1, 1)
        self.layout().addWidget(self._label, 1, 2, 1, 1)
        self.layout().addWidget(self._editor, 1, 3, 1, 1)

    @property
    def value(self):
        return self._slider.value

    def _update_editor_text(self):
        self._editor.setText('{value:.{precision}f}'.format(value=self.value, precision=self.precision))

    def update_from_slider(self):
        self._update_editor_text()
        self.value_changed.emit(self.value)

    def _update_from_editor(self):
        try:
            self._slider.index = self._slider.index_from_value(float(self._editor.text()))

            # the true slider value might stay the same, which would require a manual update of the text editor
            self._update_editor_text()

        except ValueError:
            logger.error(f'Invalid {self.full_name} value: {self._editor.text()}')
            self.reset()

    def update_default_value(self, cat: Table, object_id, translate):
        try:
            self._slider.default_value = cat.loc[object_id][self.cat_name]
        except KeyError:
            logger.warning(column_not_found_message(self.cat_name, translate))
            self._slider.default_value = self._default_value_backup

    def reset(self):
        self._slider.reset()

        # see a comment in _update_from_editor
        self._update_editor_text()

    def clear(self):
        self._editor.setText("")



import logging
import pathlib
import shutil
import yaml
from functools import wraps
from dataclasses import dataclass, asdict

import dacite
from dacite.exceptions import WrongTypeError, MissingValueError
from dictdiffer import diff, patch, swap

logger = logging.getLogger(__name__)


@dataclass
class LocalFile:
    directory: str
    filename: str = f"{__package__.split('.')[0]}.yml"
    full_name: str = "Local file"
    auto_backup: bool = True

    @property
    def path(self) -> pathlib.Path:
        return pathlib.Path(self.directory) / self.filename

    @staticmethod
    def missing_ok(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.path.exists():
                return
            func(self, *args, **kwargs)
        return wrapper

    def save(self, data: dict) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)

        if self.path.exists():
            msg = f"{self.full_name} updated (path: {self.path})"
        else:
            msg = f"{self.full_name} created (path: {self.path})"

        save_yaml(self.path, data)

        logger.info(msg)

    @missing_ok
    def backup(self) -> None:
        dst = self.path.parent / (self.filename + '.bak')
        shutil.copy(self.path, dst)
        logger.info(f'{self.full_name} backed up (path: {dst})')

    @missing_ok
    def delete(self) -> None:
        # backup the file
        if self.auto_backup:
            self.backup()

        # delete the file
        self.path.unlink()
        logger.info(f'{self.full_name} deleted (path: {self.path})')


@dataclass
class Params:
    def __post_init__(self):
        self._user_file: LocalFile | None = None

    @classmethod
    def _read(cls, filename: pathlib.Path):
        params_dict = read_yaml(filename)
        return dacite.from_dict(data_class=cls, data=params_dict, config=dacite.Config(strict=True))

    @classmethod
    def read_default_params(cls, filename: str):
        return cls._read(pathlib.Path(__file__).parent.parent / 'data' / filename)

    @classmethod
    def read_user_params(cls, file: LocalFile, default: str | None = None):
        if default is None:
            params = dacite.from_dict(data_class=cls, data={})
        else:
            params = cls.read_default_params(default)

        user_params = None

        try:
            user_params = read_yaml(file.path)
        except FileNotFoundError:
            pass
        except yaml.YAMLError:
            logger.error(f'Failed to parse `{file.path}`. The file will be overwritten.')
        else:
            # TODO: patch the user config file using dictdiffer
            try:
                user_params = dacite.from_dict(data_class=cls, data=user_params, config=dacite.Config())
            except (WrongTypeError, MissingValueError):
                user_params = None
                logger.error(f'Failed to create a dataclass from `{file.path}`. The file will be overwritten.')

        if user_params is None:
            file.backup()
        else:
            params = user_params

        params._user_file = file
        params.save()

        return params

    def replace_params(self, filename: pathlib.Path):
        try:
            params = self._read(filename)
        except Exception as e:
            logger.error(str(e))
            return None

        params._user_file = self._user_file
        return params

    def save(self, file: LocalFile | None = None):
        if file is not None:
            output_file = file
        elif self._user_file is not None:
            output_file = self._user_file
        else:
            logger.error('No output file specified')
            return

        output_file.save(data=asdict(self))

    def get_user_params_filename(self) -> str | None:
        return str(self._user_file.path.resolve()) if self._user_file is not None else None


def read_yaml(filename) -> dict:
    with open(filename, "r") as yaml_file:
        return yaml.safe_load(yaml_file)


def filter_none_values(data):
    if isinstance(data, dict):
        return {k: filter_none_values(v) for k, v in data.items() if v is not None}
    return data


def save_yaml(filename, data):
    with open(filename, 'w') as yaml_file:
        yaml.safe_dump(data, yaml_file, sort_keys=False)



from dataclasses import dataclass, field

from ..utils.params import Params


@dataclass
class Catalogue:
    filename: str | None = None
    translate: dict[str, list[str]] | None = None


@dataclass
class Data:
    dir: str = '.'
    id_pattern: str = r'\d+'
    defined_units: dict[str, str] | None = None


@dataclass
class Appearance:
    theme: str = 'light'
    antialiasing: bool = False


@dataclass
class ObjectInfo:
    show_all: bool = False
    items: list[str] | None = field(default_factory=lambda: ['ra', 'dec'])


@dataclass
class ReviewForm:
    default_checkboxes: dict[str, str] | None = None


@dataclass
class DataViewer:
    spacing: int = 5
    margins: int = 5
    label_style: dict[str, str] = field(default_factory=lambda: {'font-size': '20px'})


@dataclass
class Config(Params):
    catalogue: Catalogue = field(default_factory=lambda: Catalogue())
    data: Data = field(default_factory=lambda: Data())
    appearance: Appearance = field(default_factory=lambda: Appearance())
    object_info: ObjectInfo = field(default_factory=lambda: ObjectInfo())
    review_form: ReviewForm = field(default_factory=lambda: ReviewForm())
    data_viewer: DataViewer = field(default_factory=lambda: DataViewer())
    plugins: list[str] = field(default_factory=list)



import logging
from dataclasses import asdict
from copy import deepcopy
from functools import partial

from specutils import Spectrum1D
from astropy.nddata import StdDevUncertainty
from astropy.table import Table

import numpy as np
from scipy.ndimage import gaussian_filter1d
from astropy import units as u

import pyqtgraph as pg
from qtpy import QtCore

from ..utils import SmartSliderWithEditor
from ..utils.table_tools import column_not_found_message

from .ViewerElement import ViewerElement
from .LazyViewerElement import LazyViewerElement
from ..config import docks
from ..config.spectral_lines import SpectralLines
from ..appdata import AppData


logger = logging.getLogger(__name__)


class Spec1DItem(pg.PlotItem):
    def __init__(self, spec: Spectrum1D | None = None, lines: SpectralLines | None = None,
                 window: tuple[float, float] | None = None,
                 label_style: dict[str, str] | None = None,
                 *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.setMouseEnabled(True, True)
        self.showAxes((True, True, True, True), showValues=(True, False, True, True))
        self.hideButtons()

        self.spec = spec
        self.lines = SpectralLines() if lines is None else lines
        self.window = window

        self.label_style = {} if label_style is None else label_style

        self._flux_plot = None
        self._flux_err_plot = None

        self._default_xrange = None
        self._default_yrange = None

        # set up the spectral lines
        self._line_artists = {}
        # TODO: store colors in config
        line_color = (175.68072, 220.68924, 46.59488)
        line_pen = pg.mkPen(color=line_color, width=1)
        for line_name, lambda0 in self.lines.list.items():
            line = pg.InfiniteLine(angle=90, movable=False, pen=line_pen)
            label = pg.TextItem(text=line_name, color=line_color, anchor=(1, 1), angle=-90)
            self._line_artists[line_name] = {'line': line, 'label': label}

    @staticmethod
    def get_default_range(data: np.ndarray):
        return np.nanmin(data), np.nanmax(data)

    def get_label_height(self):
        y_min, y_max = self._default_yrange
        return y_min + 0.6 * (y_max - y_min)

    def set_spec(self, spec: Spectrum1D | None):
        self.spec = spec

        if spec is not None:
            self._default_xrange = self.get_default_range(self.spec.spectral_axis.value)
            self._default_yrange = self.get_default_range(self.spec.flux.value)

    def update_labels(self):
        self.setLabel('bottom', self.spec.spectral_axis.unit, **self.label_style)
        self.setLabel('right', self.spec.flux.unit, **self.label_style)

    def add_items(self):
        self._flux_plot = self.plot(pen='k' if pg.getConfigOption('background') == 'w' else 'w')
        self._flux_err_plot = self.plot(pen='r')

        for line_name, line_artist in self._line_artists.items():
            self.addItem(line_artist['line'], ignoreBounds=True)
            self.addItem(line_artist['label'])

    def fit_in_window(self, window):
        self.setXRange(window[0].to(self.spec.spectral_axis.unit).value,
                       window[1].to(self.spec.spectral_axis.unit).value,
                       padding=0)

    def set_line_positions(self, scale0: float = 1):
        label_height = self.get_label_height()

        for line_name, line_artist in self._line_artists.items():
            line_wave = (self.lines.list[line_name] * u.Unit(self.lines.wave_unit)).to(self.spec.spectral_axis.unit)
            line_wave = line_wave.value * scale0
            line_artist['line'].setPos(line_wave)
            line_artist['label'].setPos(QtCore.QPointF(line_wave, label_height))

        if self.window is not None:
            self.fit_in_window((scale0 * (self.window[0] + self.window[1]) / 2 - (self.window[1] - self.window[0]) / 2,
                                scale0 * (self.window[0] + self.window[1]) / 2 + (self.window[1] - self.window[0]) / 2))

    def plot_all(self):
        self._flux_plot.setData(self.spec.spectral_axis.value, self.spec.flux.value)
        if self.spec.uncertainty is not None:
            self._flux_err_plot.setData(self.spec.spectral_axis.value, self.spec.uncertainty.array)

    def display(self):
        self.update_labels()
        self.add_items()
        self.plot_all()

    def reset(self):
        if self.window is None:
            self.setXRange(*self._default_xrange, padding=0)
        self.setYRange(*self._default_yrange)

    def smooth(self, sigma: float):
        new_flux = gaussian_filter1d(self.spec.flux.value, sigma) if sigma > 0 else self.spec.flux.value

        self._default_yrange = self.get_default_range(new_flux)
        self._flux_plot.setData(self.spec.spectral_axis.value, new_flux)


class Spec1DRegion(LazyViewerElement):
    def __init__(self, lines: SpectralLines, target_line: str, cfg: docks.SpectrumRegion, **kwargs):
        super().__init__(cfg=cfg, **kwargs)

        self.line = target_line
        self.cfg = cfg

        # set up the plot
        lines = deepcopy(lines)
        lines = SpectralLines(wave_unit=lines.wave_unit, list={target_line: lines.list[target_line]})

        window_center = lines.list[target_line] * u.Unit(lines.wave_unit)
        window_size = u.Quantity(self.cfg.window_size)
        window = (window_center - window_size / 2, window_center + window_size / 2)

        self.spec_1d = Spec1DItem(lines=lines, window=window, name=self.title,
                                  label_style=self.global_config.label_style)
        self.graphics_layout.addItem(self.spec_1d)


class Spec1D(ViewerElement):
    def __init__(self, cfg: docks.Spectrum, lines: SpectralLines | None = None, **kwargs):
        super().__init__(cfg=cfg, **kwargs)

        self.lines = lines
        self.cfg = cfg

        self.lazy_widgets: list[Spec1DRegion] = []
        self.region_items: list[pg.LinearRegionItem] = []

        self.allowed_data_types = (Spectrum1D,)

        # create a redshift slider
        self._z_slider = self.create_redshift_slider(**asdict(self.cfg.redshift_slider))
        self.sliders.append(self._z_slider)

        # set up the plot
        self.spec_1d = Spec1DItem(lines=self.lines, name=self.title, label_style=self.global_config.label_style)
        self.graphics_layout.addItem(self.spec_1d)

        # create widgets zoomed on selected spectral lines
        if self.lines is not None and self.cfg.tracked_lines is not None:
            self.lazy_widgets, self.region_items = self.create_spectral_regions(self.cfg.tracked_lines, self.lines)

    def create_redshift_slider(self, **kwargs):
        redshift_slider = SmartSliderWithEditor(parameter='z', full_name='redshift', parent=self, **kwargs)
        redshift_slider.value_changed[float].connect(self._redshift_changed_action)
        return redshift_slider

    def create_spectral_regions(self, tracked_lines: dict[str, docks.SpectrumRegion], lines: SpectralLines)\
            -> tuple[list[Spec1DRegion], list[pg.LinearRegionItem]]:

        region_widgets = []
        region_items = []

        n = 0
        for line, line_cfg in tracked_lines.items():
            if line in lines.list.keys():
                spec_region = Spec1DRegion(lines=lines, target_line=line, cfg=line_cfg,
                                           title=f"{self.title} [{line}]", global_viewer_config=self.global_config,
                                           parent=self.parent())
                region_widgets.append(spec_region)

                lr = pg.LinearRegionItem()
                region_items.append(lr)

                spec_region.spec_1d.sigXRangeChanged.connect(partial(self._region_widget_changed_action, n))
                lr.sigRegionChanged.connect(partial(self._region_item_changed_action, n))

                n += 1
            else:
                logger.warning(f'Unknown spectral line: {line}')

        return region_widgets, region_items

    def _redshift_changed_action(self, redshift: float):
        self.spec_1d.set_line_positions(1 + redshift)
        for w in self.lazy_widgets:
            w.spec_1d.set_line_positions(1 + redshift)

    def _region_item_changed_action(self, n: int):
        region = self.region_items[n].getRegion()
        self.lazy_widgets[n].spec_1d.fit_in_window((region[0] * self.spec_1d.spec.spectral_axis.unit,
                                                    region[1] * self.spec_1d.spec.spectral_axis.unit))

    def _region_widget_changed_action(self, n: int):
        self.region_items[n].setRegion(self.lazy_widgets[n].spec_1d.getViewBox().viewRange()[0])

    def _load_data(self, rd: AppData):
        super()._load_data(rd=rd)
        if self.data is None:
            return

        self.spec_1d.set_spec(self.data)
        for w in self.lazy_widgets:
            w.spec_1d.set_spec(self.data)

    def display(self):
        self.spec_1d.display()
        for lr in self.region_items:
            self.spec_1d.addItem(lr)
        for w in self.lazy_widgets:
            w.spec_1d.display()

    def reset_view(self):
        self.spec_1d.reset()
        for w in self.lazy_widgets:
            w.spec_1d.reset()

        self._z_slider.reset()
        self._redshift_changed_action(self._z_slider.value)

    def clear_content(self):
        self.spec_1d.clear()
        for w in self.lazy_widgets:
            w.spec_1d.clear()

    def smooth(self, sigma: float):
        self.spec_1d.smooth(sigma)
        for w in self.lazy_widgets:
            w.spec_1d.smooth(sigma)



from dataclasses import dataclass, field

from ..utils.params import Params


@dataclass
class SpectralLines(Params):
    wave_unit: str = 'angstrom'
    list: dict[str, float] = field(default_factory=dict)



from qtpy import QtWidgets

from ..utils import AbstractWidget
from ..appdata import AppData
from ..config import config
from ..io.output import get_checkboxes


class ReviewForm(AbstractWidget):
    def __init__(self, rd: AppData, cfg: config.ReviewForm, parent=None):
        super().__init__(layout=QtWidgets.QGridLayout(), parent=parent)

        self.rd = rd
        self.cfg = cfg

        self.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred)

        # create checkboxes
        self._checkboxes = self.create_checkbox_widgets(self.cfg.default_checkboxes)

        # create a multi-line text editor for writing comments
        self._comments_widget = QtWidgets.QTextEdit(parent=self)

    def create_checkbox_widgets(self, checkboxes: dict[str, str] | None) -> dict[str, QtWidgets.QCheckBox]:
        if checkboxes is None:
            return {}

        checkbox_widgets = {}
        for i, (cname, label) in enumerate(checkboxes.items()):
            checkbox_widgets[cname] = QtWidgets.QCheckBox(label, parent=self)

        return checkbox_widgets

    def init_ui(self):
        for i, widget in enumerate(self._checkboxes.values()):
            self.layout().addWidget(widget, i + 1, 1, 1, 1)

        self.layout().addWidget(QtWidgets.QLabel('Comments:', parent=self), len(self._checkboxes) + 1, 1, 1, 1)
        self.layout().addWidget(self._comments_widget, len(self._checkboxes) + 2, 1, 1, 1)

    def dump(self):
        self.rd.df.at[self.rd.id, 'comment'] = self._comments_widget.toPlainText()
        for cname, widget in self._checkboxes.items():
            self.rd.df.at[self.rd.id, cname] = widget.isChecked()

    def load_object(self):
        self._comments_widget.setText(self.rd.df.at[self.rd.id, 'comment'])
        for cname, widget in self._checkboxes.items():
            widget.setChecked(self.rd.df.at[self.rd.id, cname])

    def load_project(self):
        for w in self._checkboxes.values():
            w.deleteLater()
        self._checkboxes = self.create_checkbox_widgets(get_checkboxes(self.rd.df, self.cfg.default_checkboxes))
        self.reset_layout()



from qtpy import QtWidgets
import pyqtgraph as pg

from ..utils import AbstractWidget
from ..config import docks, config


class LazyViewerElement(AbstractWidget):
    def __init__(self, cfg: docks.LazyViewerElement, title: str, global_viewer_config: config.DataViewer, **kwargs):
        super().__init__(layout=QtWidgets.QGridLayout(), **kwargs)

        self.cfg = cfg
        self.title = title
        self.global_config = global_viewer_config

        # create a central widget
        self.graphics_view = pg.GraphicsView()
        self.graphics_layout = pg.GraphicsLayout()
        self.graphics_view.setCentralItem(self.graphics_layout)

        self.set_geometry(spacing=global_viewer_config.spacing, margins=global_viewer_config.margins)

    def set_geometry(self, spacing: int, margins: int | tuple[int, int, int, int]):
        super().set_geometry(spacing=spacing, margins=margins)

        self.graphics_layout.setSpacing(spacing)
        self.graphics_layout.setContentsMargins(0, 0, 5, 5)

    def init_ui(self):
        self.layout().addWidget(self.graphics_view, 1, 1, 1, 1)



import logging

# from astropy.coordinates import SkyCoord
from qtpy import QtWidgets, QtCore

from ..utils import AbstractWidget
from ..utils import column_not_found_message
from ..appdata import AppData
from ..config import config


logger = logging.getLogger(__name__)


class ObjectInfo(AbstractWidget):
    def __init__(self, rd: AppData, cfg: config.ObjectInfo, parent=None):
        super().__init__(layout=QtWidgets.QGridLayout(), parent=parent)

        self.rd = rd
        self.cfg = cfg

        self.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred)

        self._table = QtWidgets.QTableWidget()
        self._table.setColumnCount(2)
        self._table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self._table.horizontalHeader().hide()
        self._table.horizontalHeader().setStretchLastSection(True)
        self._table.horizontalHeader().setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeToContents)
        # self._table.setHorizontalHeaderLabels(('key', 'value'))
        self._table.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Fixed)

        # display information about the object
        self._table_items = self.create_table_items(self.cfg.items)
        self.set_items()

    @staticmethod
    def create_table_items(items) -> list[tuple[QtWidgets.QTableWidgetItem, QtWidgets.QTableWidgetItem]]:
        if items is None:
            return []

        table_items = []
        for cname in items:
            cname_item = QtWidgets.QTableWidgetItem(cname)
            value_item = QtWidgets.QTableWidgetItem('')

            table_items.append((cname_item, value_item))

        return table_items

    def set_items(self):
        self._table.setRowCount(len(self._table_items))
        for i, row in enumerate(self._table_items):
            self._table.setItem(i, 0, row[0])
            self._table.setItem(i, 1, row[1])

    def update_items(self):
        if self.cfg.show_all and self.rd.cat is not None:
            self._table_items = self.create_table_items(self.rd.cat.colnames)
        else:
            self._table_items = self.create_table_items(self.cfg.items)
        self.set_items()

    def init_ui(self):
        self.layout().addWidget(self._table, 1, 1, 1, 1)

    def load_object(self):
        for row in self._table_items:
            cname = row[0].text()
            try:
                row[1].setText(str(self.rd.cat.loc[self.rd.id][cname]))
            except KeyError:
                if cname in self.rd.cat.colnames:
                    logger.warning('Object not found in the catalogue (ID: {})'.format(self.rd.id))
                else:
                    logger.warning(column_not_found_message(cname, self.rd.config.catalogue.translate))
                row[1].setText('')

        # if 'ra' in self._cat.colnames and 'dec' in self._cat.colnames:
        #     c = SkyCoord(ra=self._cat['ra'][self._j], dec=self._cat['dec'][self._j], frame='icrs', unit='deg')
        #     ra, dec = c.to_string('hmsdms').split(' ')
        #     self.ra_label.setText("RA: {}".format(ra))
        #     self.dec_label.setText("Dec: {}".format(dec))



import importlib

from qtpy import QtWidgets

from pyqtgraph.dockarea.Dock import Dock
from pyqtgraph.dockarea.DockArea import DockArea

from ..utils import AbstractWidget
from .ViewerElement import ViewerElement
from .LazyViewerElement import LazyViewerElement
from .Image2D import Image2D
from .Spec1D import Spec1D

from ..appdata import AppData
from ..io.viewer_data import add_enabled_aliases


class DataViewer(AbstractWidget):
    def __init__(self, rd: AppData, parent=None):
        super().__init__(layout=QtWidgets.QGridLayout(), parent=parent)

        self.rd = rd

        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)

        # register units
        if self.rd.config.data.defined_units is not None:
            add_enabled_aliases(self.rd.config.data.defined_units)

        plugins = self.rd.config.plugins
        self._plugins = [importlib.import_module("specvizitor.plugins." + plugin_name).Plugin()
                         for plugin_name in plugins] if plugins is not None else []

        self.set_geometry(spacing=0, margins=0)

        self.dock_area = DockArea()
        self.added_docks: list[str] = []

        self.data_widgets: dict[str, ViewerElement] = {}
        self.docks: dict[str, Dock] = {}

        self.create_widgets()
        self.create_docks()
        self.add_docks()

        try:
            # set `extra` to None to catch an exception (KeyError) when adding extra docks not mentioned in `state`
            self.dock_area.restoreState(self.rd.cache.dock_state, missing='ignore', extra=None)
        except (KeyError, ValueError, TypeError):
            self.reset_dock_state()

    @property
    def widgets(self) -> list[LazyViewerElement]:
        lazy_widgets = []
        for w in self.data_widgets.values():
            lazy_widgets.extend(w.lazy_widgets)

        return list(self.data_widgets.values()) + lazy_widgets

    def init_ui(self):
        self.layout().addWidget(self.dock_area, 1, 1, 1, 1)

    def create_widgets(self):
        # delete previously created widgets
        if self.widgets:
            for w in self.widgets:
                w.deleteLater()

        widgets = {}

        # create widgets for images (e.g. image cutouts, 2D spectra)
        if self.rd.docks.images is not None:
            for name, image_cfg in self.rd.docks.images.items():
                widgets[name] = Image2D(cfg=image_cfg, title=name, global_viewer_config=self.rd.config.data_viewer,
                                        parent=self)

        # create widgets for 1D spectra
        if self.rd.docks.spectra is not None:
            for name, spec_cfg in self.rd.docks.spectra.items():
                widgets[name] = Spec1D(lines=self.rd.lines, cfg=spec_cfg, title=name,
                                       global_viewer_config=self.rd.config.data_viewer, parent=self)

        for w in widgets.values():
            w.init_ui()

        self.data_widgets = widgets

    def create_docks(self):
        docks = {}
        for widget in self.widgets:
            docks[widget.title] = Dock(widget.title, widget=widget)

        self.docks = docks

    def add_dock(self, widget: LazyViewerElement):
        self.dock_area.addDock(dock=self.docks[widget.title],
                               position=widget.cfg.position if widget.cfg.position is not None else 'bottom',
                               relativeTo=widget.cfg.relative_to if widget.cfg.relative_to in self.added_docks else None)
        self.added_docks.append(widget.title)

    def add_docks(self):
        self.added_docks = []

        for widget in self.widgets:
            if widget.cfg.visible:
                self.add_dock(widget)

    def update_dock_titles(self):
        for w in self.data_widgets.values():
            if w.filename is not None and w.data is not None:
                self.docks[w.title].setTitle(w.filename.name)
                for lazy_widget in w.lazy_widgets:
                    self.docks[lazy_widget.title].setTitle(w.filename.name)
            else:
                self.docks[w.title].setTitle(w.title)
                for lazy_widget in w.lazy_widgets:
                    self.docks[lazy_widget.title].setTitle(lazy_widget.title)

    def reset_dock_state(self):
        for dock_name in self.added_docks:
            self.docks[dock_name].close()
            self.docks[dock_name].deleteLater()
        self.create_docks()
        self.add_docks()
        self.update_dock_titles()

    def save_dock_state(self):
        self.rd.cache.dock_state = self.dock_area.saveState()
        self.rd.cache.save()

    def load_object(self):
        self.save_dock_state()

        # load the object to the widget
        for w in self.data_widgets.values():
            w.load_object(rd=self.rd)

        # update the title of the dock
        self.update_dock_titles()

        for plugin in self._plugins:
            plugin.link(self.data_widgets, label_style=self.rd.config.data_viewer.label_style)

    def reset_view(self):
        for w in self.data_widgets.values():
            if w.data is not None:
                w.reset_view()

    def take_screenshot(self, filename: str):
        self.grab().save(filename)



import pathlib
import argparse
import sys
import logging
from importlib.metadata import version
from dataclasses import asdict

from platformdirs import user_config_dir, user_cache_dir

import pyqtgraph as pg
import qtpy.compat
from qtpy import QtGui, QtWidgets, QtCore
from qtpy.QtCore import Slot

import qdarktheme

from .appdata import AppData
from .config import Config, Docks, SpectralLines, Cache
from .menu import NewFile, Settings
from .widgets import (AbstractWidget, DataViewer, ControlBar, QuickSearch, ObjectInfo, ReviewForm)
from .utils.logs import LogMessageBox
from .utils.params import LocalFile, save_yaml


logger = logging.getLogger(__name__)


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, appdata: AppData, parent=None):
        self.rd = appdata

        super().__init__(parent)

        self.was_maximized: bool
        self.setFocusPolicy(QtCore.Qt.StrongFocus)

        self.setWindowTitle('Specvizitor')  # set the title of the main window
        # self.setWindowIcon(QtGui.QIcon('logo2_2.png'))

        # create a central widget
        self.central_widget = DataViewer(self.rd, parent=self)

        # add a menu bar
        self._init_menu()

        # add a status bar
        # self.statusBar().showMessage("Message in the statusbar")

        self.widgets: list[AbstractWidget] = []

        # add a central widget
        self.setCentralWidget(self.central_widget)

        # create a control panel
        self.control_bar = ControlBar(self.rd, parent=self)
        self.control_bar.setObjectName('Control Bar')
        self.addToolBar(QtCore.Qt.TopToolBarArea, self.control_bar)

        # create a quick search widget
        self.quick_search = QuickSearch(rd=self.rd, parent=self)
        self.quick_search_dock = QtWidgets.QDockWidget('Quick Search', self)
        self.quick_search_dock.setObjectName('Quick Search')
        self.quick_search_dock.setWidget(self.quick_search)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.quick_search_dock)

        # create a widget displaying information about the object
        self.object_info = ObjectInfo(self.rd, cfg=self.rd.config.object_info, parent=self)
        self.object_info_dock = QtWidgets.QDockWidget('Object Information', self)
        self.object_info_dock.setObjectName('Object Information')
        self.object_info_dock.setWidget(self.object_info)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.object_info_dock)

        # create a widget for writing comments
        self.review_form = ReviewForm(self.rd, cfg=self.rd.config.review_form, parent=self)
        self.review_form_dock = QtWidgets.QDockWidget('Review Form', self)
        self.review_form_dock.setObjectName('Review Form')
        self.review_form_dock.setWidget(self.review_form)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.review_form_dock)

        # connect signals from the control panel and the quick search window to the slots of the central widget
        self.control_bar.object_selected.connect(self.load_object)
        self.quick_search.object_selected.connect(self.load_object)
        self.control_bar.screenshot_button_clicked.connect(self.central_widget.take_screenshot)
        self.control_bar.reset_view_button_clicked.connect(self.central_widget.reset_view)
        self.control_bar.reset_dock_state_button_clicked.connect(self.central_widget.reset_dock_state)

        for w in (self.central_widget, self.control_bar, self.quick_search, self.object_info, self.review_form):
            self.widgets.append(w)

        self._init_ui()

        # read cache and try to load the last active project
        if self.rd.cache.last_inspection_file:
            self.open_file(self.rd.cache.last_inspection_file)

        settings = QtCore.QSettings()
        if settings.value("geometry") is None or settings.value("windowState") is None:
            self.showMaximized()
            self.setFocus()
        else:
            self.restoreGeometry(settings.value("geometry"))
            self.restoreState(settings.value("windowState"))

    def _init_menu(self):
        self._menu = self.menuBar()

        self._file = self._menu.addMenu("&File")

        self._new_file = QtWidgets.QAction("&New...")
        self._new_file.triggered.connect(self._new_file_action)
        self._new_file.setShortcut(QtGui.QKeySequence('Ctrl+N'))
        self._file.addAction(self._new_file)

        self._open_file = QtWidgets.QAction("&Open...")
        self._open_file.triggered.connect(self._open_file_action)
        self._open_file.setShortcut(QtGui.QKeySequence('Ctrl+O'))
        self._file.addAction(self._open_file)

        self._file.addSeparator()

        self._save = QtWidgets.QAction("&Save...")
        self._save.triggered.connect(self._save_action)
        self._save.setShortcut(QtGui.QKeySequence('Ctrl+S'))
        self._save.setEnabled(False)
        self._file.addAction(self._save)

        self._save_as = QtWidgets.QAction("Save As...")
        self._save_as.triggered.connect(self._save_as_action)
        self._save_as.setShortcut(QtGui.QKeySequence('Shift+Ctrl+S'))
        self._save_as.setEnabled(False)
        self._file.addAction(self._save_as)

        self._export = QtWidgets.QAction("&Export...")
        self._export.triggered.connect(self._export_action)
        self._export.setEnabled(False)
        self._file.addAction(self._export)

        self._file.addSeparator()

        self._quit = QtWidgets.QAction("&Quit...")
        self._quit.triggered.connect(self._exit_action)
        self._quit.setShortcut(QtGui.QKeySequence('Ctrl+Q'))
        self._file.addAction(self._quit)

        self._view = self._menu.addMenu("&View")

        self._reset_view = QtWidgets.QAction("Reset View")
        self._reset_view.setShortcut('F5')
        self._reset_view.triggered.connect(self.central_widget.reset_view)
        self._view.addAction(self._reset_view)

        self._reset_dock_state = QtWidgets.QAction("Reset Dock State")
        self._reset_dock_state.triggered.connect(self.central_widget.reset_dock_state)
        self._view.addAction(self._reset_dock_state)

        self._view.addSeparator()

        self._backup_dock_configuration = QtWidgets.QAction("Backup...")
        self._backup_dock_configuration.triggered.connect(self._backup_dock_configuration_action)
        self._view.addAction(self._backup_dock_configuration)

        self._restore_dock_configuration = QtWidgets.QAction("Restore...")
        self._restore_dock_configuration.triggered.connect(self._restore_dock_configuration_action)
        self._view.addAction(self._restore_dock_configuration)

        self._view.addSeparator()

        self._fullscreen = QtWidgets.QAction("Fullscreen")
        self._fullscreen.triggered.connect(lambda:
                                           self._exit_fullscreen() if self.isFullScreen() else self._enter_fullscreen())
        self._fullscreen.setShortcut('F11')
        self._view.addAction(self._fullscreen)

        self._shortcut_fullscreen = QtWidgets.QShortcut('Esc', self)
        self._shortcut_fullscreen.activated.connect(lambda: self._exit_fullscreen() if self.isFullScreen() else None)

        self._tools = self._menu.addMenu("&Tools")
        self._settings = QtWidgets.QAction("Se&ttings...")
        self._settings.triggered.connect(self._settings_action)
        self._tools.addAction(self._settings)

        self._help = self._menu.addMenu("&Help")
        self._about = QtWidgets.QAction("&About...")
        self._about.triggered.connect(self._about_action)
        self._help.addAction(self._about)

    def _init_ui(self):
        for widget in self.widgets:
            widget.init_ui()

    def _new_file_action(self):
        """ Create a new inspection file via the NewFile dialog.
        """
        dialog = NewFile(self.rd, parent=self)
        if dialog.exec():
            self.rd.cache.last_object_index = 0
            self.load_project()

    def _open_file_action(self):
        """ Open an existing inspection file via QFileDialog.
        """
        path = qtpy.compat.getopenfilename(self, caption='Open Inspection File', filters='CSV Files (*.csv)')[0]
        if path:
            self.rd.cache.last_object_index = 0
            self.open_file(path)

    def open_file(self, path: str):
        """ Load inspection data from an existing inspection file.
        @param path: path to the inspection file
        """
        if pathlib.Path(path).exists():
            self.rd.output_path = pathlib.Path(path)
            self.rd.read()
            self.load_project()
        else:
            logger.warning('Inspection file not found (path: {})'.format(path))

    def load_project(self):
        """ Update the state of the main window and activate the central widget after loading inspection data.
        """
        for w in (self._save, self._save_as, self._export):
            w.setEnabled(True)

        self.review_form.load_project()
        self.object_info.update_items()
        for widget in self.widgets:
            widget.setEnabled(True)

        # cache the inspection file name
        self.rd.cache.last_inspection_file = str(self.rd.output_path)
        self.rd.cache.save()

        # try to display the object with an index stored in cache
        j = self.rd.cache.last_object_index
        if j and j < self.rd.n_objects:
            self.load_object(int(j))
        else:
            self.load_object(0)

    @Slot(int)
    def load_object(self, j: int):
        """ Load a new object to the central widget.
        @param j: the index of the object to display
        """
        if self.rd.j is not None:
            # update the application data from widgets
            self.review_form.dump()

            self.rd.save()  # auto-save

        self.rd.j = j

        for widget in (self.central_widget, self.control_bar, self.object_info, self.review_form):
            widget.load_object()

        self.setWindowTitle(
            f'{self.rd.output_path.name} – ID {self.rd.id} [#{self.rd.j + 1}/{self.rd.n_objects}] – Specvizitor')

        # cache the index of the object
        # TODO: cache the ID instead of the index
        self.rd.cache.last_object_index = j
        self.rd.cache.save()

    def _save_action(self):
        """ Instead of saving inspection results, display a message saying that the auto-save mode is enabled.
        """
        msg = 'The project data is saved automatically'
        if self.rd.output_path is not None:
            msg += ' to {}'.format(self.rd.output_path)
        LogMessageBox(logging.INFO, msg, parent=self)

    def _save_as_action(self):
        LogMessageBox(logging.INFO, 'Not implemented', parent=self)

    def _export_action(self):
        LogMessageBox(logging.INFO, 'Not implemented', parent=self)

    def _exit_action(self):
        if self.rd.df is not None:
            self.rd.save()  # auto-save
        self.central_widget.save_dock_state()  # save the dock state of the central widget

        # save the state and geometry of the main window
        settings = QtCore.QSettings()
        settings.setValue('geometry', self.saveGeometry())
        settings.setValue('windowState', self.saveState())

        self.close()
        logger.info("Application closed")

    def _restore_dock_configuration_action(self):
        path = qtpy.compat.getopenfilename(self, caption='Open Dock Configuration',
                                           basedir=str(pathlib.Path()),
                                           filters='YAML Files (*.yml)')[0]
        if path:
            new_docks = self.rd.docks.replace_params(pathlib.Path(path))
            if new_docks is None:
                logger.error('Failed to restore dock configuration')
            else:
                self.rd.docks = new_docks
                self.rd.docks.save()

                self.central_widget.create_widgets()
                self.central_widget.reset_dock_state()

                if self.rd.df is not None:
                    self.central_widget.load_object()

                logger.info('Dock configuration restored')

    def _backup_dock_configuration_action(self):
        path = qtpy.compat.getsavefilename(self, caption='Save Dock Configuration',
                                           basedir=str(pathlib.Path() / 'docks.yml'),
                                           filters='YAML Files (*.yml)')[0]

        if path:
            save_yaml(path, asdict(self.rd.docks))
            logger.info('Dock configuration saved')

    def _enter_fullscreen(self):
        self.was_maximized = True if self.isMaximized() else False
        self.showFullScreen()

    def _exit_fullscreen(self):
        self.showNormal()
        if self.was_maximized:
            self.showMaximized()

    def _settings_action(self):
        dialog = Settings(self.rd, parent=self)
        if dialog.exec():
            self.object_info.update_items()
            if self.rd.df is not None:
                for widget in (self.central_widget, self.object_info):
                    widget.load_object()

    def _about_action(self):
        QtWidgets.QMessageBox.about(self, "About Specvizitor", "Specvizitor v{}".format(version('specvizitor')))

    def closeEvent(self, _):
        self._exit_action()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument('--purge', action='store_true')

    args = parser.parse_args()

    # logging configuration
    level = logging.INFO if args.verbose else logging.WARNING
    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', level=level)

    user_files: dict[str, LocalFile] = {
        'config': LocalFile(user_config_dir('specvizitor'), full_name='Settings file'),
        'docks': LocalFile(user_config_dir('specvizitor'), filename='docks.yml', full_name='Dock configuration file'),
        'lines': LocalFile(user_config_dir('specvizitor'), filename='lines.yml', full_name='List of spectral lines'),
        'cache': LocalFile(user_cache_dir('specvizitor'), full_name='Cache file', auto_backup=False)
    }

    if args.purge:
        for f in user_files.values():
            f.delete()

    # initialize the app data
    appdata = AppData(config=Config.read_user_params(user_files['config'], default='default_config.yml'),
                      docks=Docks.read_user_params(user_files['docks'], default='default_docks.yml'),
                      lines=SpectralLines.read_user_params(user_files['lines'], default='default_lines.yml'),
                      cache=Cache.read_user_params(user_files['cache']))

    # pyqtgraph configuration
    pg.setConfigOption('background', 'w')
    pg.setConfigOption('foreground', 'k')
    pg.setConfigOption('imageAxisOrder', 'row-major')
    pg.setConfigOption('antialias', appdata.config.appearance.antialiasing)

    # start the application
    app = QtWidgets.QApplication(sys.argv)
    app.setOrganizationName('FRESCO')
    app.setApplicationName('Specvizitor')
    logger.info("Application started")

    # set up the theme
    qdarktheme.setup_theme(appdata.config.appearance.theme)
    if appdata.config.appearance.theme == 'dark':
        pg.setConfigOption('background', "#1d2023")
        pg.setConfigOption('foreground', '#eff0f1')
    else:
        pg.setConfigOption('background', "w")
        pg.setConfigOption('foreground', 'k')

    # initialize the main window
    window = MainWindow(appdata=appdata)
    window.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



import logging

from platformdirs import user_config_dir, user_cache_dir
from qtpy import QtWidgets

from ..appdata import AppData
from ..io.catalogue import load_cat, create_cat, cat_browser
from ..io.viewer_data import data_browser
from ..utils.logs import qlog


logger = logging.getLogger(__name__)


class Settings(QtWidgets.QDialog):
    def __init__(self, rd: AppData, parent=None):
        self.rd = rd

        super().__init__(parent)
        self.setWindowTitle("Settings [Beta]")

        layout = QtWidgets.QVBoxLayout(self)
        layout.setSpacing(20)

        self._browsers = {
            'data': data_browser(self.rd.config.data.dir, self),
            'cat': cat_browser(self.rd.config.catalogue.filename, self)
        }

        for b in self._browsers.values():
            layout.addWidget(b)

        # add a horizontal separator
        self.separator = QtWidgets.QFrame(self)
        self.separator.setFrameShape(QtWidgets.QFrame.HLine)
        layout.addWidget(self.separator)

        self._info_label = QtWidgets.QLabel(f"GUI Configuration: {user_config_dir('specvizitor')}\n\n"
                                            f"Cache: {user_cache_dir('specvizitor')}", parent=self)
        layout.addWidget(self._info_label)

        # add OK/Cancel buttons
        self._button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel, parent=self)
        layout.addWidget(self._button_box)
        self._button_box.accepted.connect(self.accept)
        self._button_box.rejected.connect(self.reject)

        self.setLayout(layout)

    @qlog
    def validate(self) -> bool:
        for name, b in self._browsers.items():
            if (name == 'data' and not b.is_filled(verbose=True)) or not b.exists(verbose=True):
                return False
        return True

    def accept(self):
        if not self.validate():
            return

        if self._browsers['cat'].is_filled():
            cat = load_cat(self._browsers['cat'].path, translate=self.rd.config.catalogue.translate)
            if not cat:
                return

            self.rd.cat = cat
            self.rd.config.catalogue.filename = self._browsers['cat'].path

        else:
            self.rd.config.catalogue.filename = None
            if self.rd.df is not None:
                self.rd.cat = create_cat(self.rd.df.index.values)

        self.rd.config.data.dir = self._browsers['data'].path
        self.rd.config.save()

        super().accept()



import logging
from qtpy import QtWidgets, QtCore

from ..utils import AbstractWidget
from ..appdata import AppData

logger = logging.getLogger(__name__)


class QuickSearch(AbstractWidget):
    object_selected = QtCore.Signal(int)

    def __init__(self, rd: AppData, parent=None):
        super().__init__(layout=QtWidgets.QGridLayout(), parent=parent)

        self.rd = rd

        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)

        # create the `Go to ID` button
        self._go_to_id_button = QtWidgets.QPushButton(self)
        self._go_to_id_button.setText('Go to ID')
        self._go_to_id_button.setFixedWidth(110)
        self._go_to_id_button.clicked.connect(self.go_to_id)

        self._id_field = QtWidgets.QLineEdit(self)
        self._id_field.returnPressed.connect(self.go_to_id)

        # create the `Go to index` button
        self._go_to_index_button = QtWidgets.QPushButton(self)
        self._go_to_index_button.setText('Go to #')
        self._go_to_index_button.setFixedWidth(110)
        self._go_to_index_button.clicked.connect(self.go_to_index)

        self._index_field = QtWidgets.QLineEdit(self)
        self._index_field.returnPressed.connect(self.go_to_index)

    def init_ui(self):
        self.layout().addWidget(self._go_to_id_button, 1, 1, 1, 1)
        self.layout().addWidget(self._id_field, 1, 2, 1, 1)
        self.layout().addWidget(self._go_to_index_button, 2, 1, 1, 1)
        self.layout().addWidget(self._index_field, 2, 2, 1, 1)

    def go_to_id(self):
        text = self._id_field.text()
        self._id_field.clear()

        try:
            id_upd = int(text)
        except ValueError:
            logger.error(f'Invalid ID: {text}')
            return

        if id_upd in self.rd.df.index:
            self.parent().setFocus()
            self.object_selected.emit(self.rd.df.index.get_loc(id_upd))
        else:
            logger.error(f'ID `{text}` not found')
            return

    def go_to_index(self):
        text = self._index_field.text()
        self._index_field.clear()

        try:
            index_upd = int(text)
        except ValueError:
            logger.error(f'Invalid index: {text}')
            return

        if 0 < index_upd <= self.rd.n_objects:
            self.parent().setFocus()
            self.object_selected.emit(index_upd - 1)
        else:
            logger.error(f'Index `{text}` out of range')
            return



import logging
import pathlib
import re
from dataclasses import dataclass
import abc
import warnings

from astropy.io import fits
from astropy.table import Table
import astropy.units as u
from astropy.io.registry import IORegistryError
from astropy.units.core import UnitConversionError
from astropy.utils.exceptions import AstropyWarning
from PIL import Image, ImageOps, UnidentifiedImageError
from specutils import Spectrum1D

import numpy as np

from ..utils import FileBrowser

logger = logging.getLogger(__name__)


@dataclass
class BaseLoader(abc.ABC):
    name: str

    @abc.abstractmethod
    def load(self, filename: pathlib.Path, **kwargs):
        pass

    def raise_error(self, e):
        logger.error(f'{type(self).__name__}: {e}')


@dataclass
class GenericFITSLoader(BaseLoader):
    name: str = 'generic_fits'

    def load(self, filename: pathlib.Path, extname: str = None, extver: str = None, **kwargs):

        try:
            hdul = fits.open(filename, **kwargs)
        except Exception as e:
            self.raise_error(e)
            return None, None

        if extname is not None and extver is None:
            index = extname
        elif extname is not None and extver is not None:
            index = (extname, extver)
        else:
            index = 1

        try:
            hdu = hdul[index]
        except KeyError:
            self.raise_error(f'Extension `{index}` not found (filename: {filename.name})')
            return None, None

        data = hdu.data
        meta = hdu.header

        if meta['XTENSION'] in ('TABLE', 'BINTABLE'):
            data = Table(data)

        return data, meta


@dataclass
class PILLoader(BaseLoader):
    name: str = 'pil'

    def load(self, filename: pathlib.Path, **kwargs):
        try:
            image = Image.open(filename, **kwargs)
        except Exception as e:
            self.raise_error(e)
            return None, None

        image = ImageOps.flip(image)

        return np.array(image), image.info


@dataclass
class SpecutilsLoader(BaseLoader):
    name: str = 'specutils'

    def load(self, filename: pathlib.Path, **kwargs):
        try:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', AstropyWarning)
                spec: Spectrum1D = Spectrum1D.read(filename, **kwargs)
        except Exception as e:
            self.raise_error(e)
            return None, None

        try:
            # specutils treats "pix" as a valid spectral axis unit
            spec.spectral_axis.unit.to(u.Unit('AA'))
        except UnitConversionError:
            self.raise_error(f'Invalid spectral axis unit: {spec.spectral_axis.unit}')
            return None, None

        return spec, spec.meta


def load(loader_name: str | None, filename: pathlib.Path, widget_name: str, **kwargs):
    registered_loaders: dict[str, BaseLoader] =\
        {loader.name: loader for loader in (GenericFITSLoader(), PILLoader(), SpecutilsLoader())}

    allowed_loader_names = ('auto',) + tuple(loader.name for loader in registered_loaders.values())
    if loader_name not in allowed_loader_names:
        logger.error(f'Unknown loader type: `{loader_name}` (widget: {widget_name}).'
                     f'Available loaders: {allowed_loader_names}')
        return None, None

    if loader_name == 'auto':
        if filename.suffix == '.fits':
            loader_name = 'generic_fits'
        else:
            loader_name = 'pil'

    return registered_loaders[loader_name].load(filename, **kwargs)


def add_enabled_aliases(units: dict[str, str]):
    u.add_enabled_aliases({alias: u.Unit(unit) for alias, unit in units.items()})


def get_filename(directory, pattern: str, object_id) -> pathlib.Path | None:
    """ Find a file for a given pattern and a given object ID. If more than one file name is matched to the pattern,
    return the first item from an alphabetically ordered list of matched file names.

    @param directory: the directory where the search is performed
    @param pattern: the pattern used to match file names
    @param object_id: the object ID
    @return: the file name matched to the pattern
    """

    filenames = sorted(pathlib.Path(directory).glob('*'))

    # match to the pattern
    matched_to_pattern = [p for p in filenames if re.search(pattern, str(p))]

    # match to the ID
    if isinstance(object_id, int):
        # make sure that we don't match e.g. '1123' or '1234' to '123' (but match '0123')
        matched_ids = [re.findall(r'\d*{}\d*'.format(object_id), str(p)) for p in matched_to_pattern]
        matched = [p for i, p in enumerate(matched_to_pattern)
                   if matched_ids[i] and str(max(matched_ids[i], key=len)).lstrip('0') == str(object_id)]
    else:
        matched = [p for p in matched_to_pattern if re.search(f'{object_id}', str(p))]

    if matched:
        return matched[0]
    else:
        return


def get_id_from_filename(filename, pattern: str) -> str | None:
    """ Extract the object ID from a file name using a pattern. If more than one ID is matched to the pattern, return
    the longest match (a typical case for integer IDs).

    @param filename: the file name to parse
    @param pattern: the pattern (a regular expression) used to match the ID
    @return: the matched ID
    """

    try:
        matches: list[str] = re.findall(pattern, pathlib.Path(filename).name)
    except re.error:
        return

    if matches:
        return max(matches, key=len)
    else:
        return


def get_ids_from_dir(directory, id_pattern: str) -> np.ndarray | None:
    """ Extract IDs from a directory using a pattern.
    @param directory: the directory where the search for IDs is performed
    @param id_pattern: the pattern (a regular expression) used to match IDs
    @return: a list of matched IDs
    """

    data_files = sorted(pathlib.Path(directory).glob('*'))  # includes subdirectories, if any
    ids = [get_id_from_filename(p, id_pattern) for p in data_files]
    ids = np.array([i for i in ids if i is not None])

    try:
        # convert IDs to int
        ids = ids.astype(np.int64)
    except ValueError:
        pass

    # remove ID duplicates
    ids = np.unique(ids)

    if not ids.size:
        logger.error('No IDs retrieved from the data directory')
        return

    return ids


def data_browser(default_path, parent) -> FileBrowser:
    return FileBrowser(title='Data Source:', mode=FileBrowser.OpenDirectory, default_path=default_path, parent=parent)



from qtpy import QtWidgets


def get_widgets(layout: QtWidgets.QLayout) -> list[QtWidgets.QWidget]:
    widgets = []

    index = layout.count() - 1
    while index >= 0:
        widget, sub_layout = layout.itemAt(index).widget(), layout.itemAt(index).layout()
        if widget is not None:
            widgets.append(widget)
        elif layout is not None:
            widgets.extend(get_widgets(sub_layout))
        index -= 1

    return widgets



import logging

import numpy as np
import pyqtgraph as pg
from astropy.visualization import ZScaleInterval
from scipy.ndimage import gaussian_filter

from pgcolorbar.colorlegend import ColorLegendItem

from .ViewerElement import ViewerElement
from ..config import docks
from ..appdata import AppData


logger = logging.getLogger(__name__)


class Image2D(ViewerElement):
    def __init__(self, cfg: docks.Image, **kwargs):
        super().__init__(cfg=cfg, **kwargs)

        self.cfg = cfg

        self.allowed_data_types = (np.ndarray,)

        # set up the color map
        self._cmap = pg.colormap.get('viridis')

        # create an image item
        self.image_2d = pg.ImageItem()
        self.image_2d.setLookupTable(self._cmap.getLookupTable())

        # create an image container
        if self.cfg.container == 'PlotItem':
            # create a plot item
            self.container = pg.PlotItem(name=self.title)
            # self.container.hideAxis('left')
            self.container.showAxes((False, False, False, True), showValues=(False, False, False, True))
            self.container.hideButtons()

            # add a border to the image
            self.image_2d.setBorder('k')
        else:
            # create a view box
            self.container = pg.ViewBox()

        # add the image to the container
        self.container.addItem(self.image_2d)

        # lock the aspect ratio
        self.container.setAspectLocked(True)

        # add the container to the layout
        self.graphics_layout.addItem(self.container, 0, 0)

        # create a color bar
        self._cbar = ColorLegendItem(imageItem=self.image_2d, showHistogram=True, histHeightPercentile=99.0)
        self._cbar.setVisible(self.cfg.color_bar.visible)

        # add the color bar to the layout
        self.graphics_layout.addItem(self._cbar, 0, 1)

    def _load_data(self, rd: AppData):
        super()._load_data(rd=rd)
        if self.data is None:
            return

        # rotate the image
        self.rotate(self.cfg.rotate)

        # scale the data points
        self.scale(self.cfg.scale)

    def display(self):
        self.image_2d.setImage(self.data)

    def reset_view(self):
        # TODO: allow to choose between min/max and zscale?
        self._cbar.setLevels(ZScaleInterval().get_limits(self.data))
        self.container.autoRange(padding=0)

    def clear_content(self):
        self.image_2d.clear()

    def smooth(self, sigma: float):
        self.image_2d.setImage(gaussian_filter(self.data, sigma) if sigma > 0 else self.data)

    def rotate(self, angle: int):
        self.data = np.rot90(self.data, k=angle // 90)

    def scale(self, scale: float):
        self.data *= scale



from qtpy import QtGui

from ..widgets.ViewerElement import ViewerElement


class Plugin:
    # TODO: create BasePlugin
    @staticmethod
    def link(widgets: dict[str, ViewerElement], label_style: dict[str, str] | None = None):

        try:
            spec_1d, spec_2d = widgets["Spectrum 1D"], widgets["Spectrum 2D"]
        except KeyError:
            return

        if any(w.data is None for w in (spec_1d, spec_2d)):
            return

        # set x-axis transformation for the 2D spectrum plot
        dlam = spec_2d.meta['CD1_1'] * 1e4
        crval = spec_2d.meta['CRVAL1'] * 1e4
        crpix = spec_2d.meta['CRPIX1']

        qtransform = QtGui.QTransform().translate(crval - dlam * crpix, 0).scale(dlam, 1)

        spec_2d.image_2d.setTransform(qtransform)
        spec_2d.container.setAspectLocked(True, 1 / dlam)
        if label_style is None:
            label_style = {}
        spec_2d.container.setLabel('bottom', spec_1d.spec_1d.spec.spectral_axis.unit, **label_style)
        spec_2d.container.setXLink(spec_1d.title)  # link the x-axis range



from dataclasses import dataclass, field
from typing import Any

from ..utils.params import Params


@dataclass
class Slider:
    visible: bool = True

    min_value: float = 0
    max_value: float = 100
    step: float = 1
    default_value: float = 0

    catalogue_name: str | None = None
    show_text_editor: bool = False
    text_editor_precision: int = 6


@dataclass
class ColorBar:
    visible: bool = True


@dataclass
class LazyViewerElement:
    visible: bool = True

    position: str | None = None
    relative_to: str | None = None


@dataclass
class ViewerElement(LazyViewerElement):
    filename_keyword: str | None = None
    data_loader: str = 'auto'
    data_loader_params: dict[str, Any] | None = None
    smoothing_slider: Slider = field(default_factory=lambda: Slider(max_value=3, step=0.05))


@dataclass
class Image(ViewerElement):
    rotate: int = 0
    scale: float = 1
    container: str = 'ViewBox'
    color_bar: ColorBar = field(default_factory=lambda: ColorBar())


@dataclass
class SpectrumRegion(LazyViewerElement):
    window_size: str = '200 Angstrom'


@dataclass
class Spectrum(ViewerElement):
    data_loader: str = 'specutils'
    redshift_slider: Slider = field(default_factory=lambda: Slider(max_value=10, step=1e-4))
    tracked_lines: dict[str, SpectrumRegion] | None = None


@dataclass
class Docks(Params):
    images: dict[str, Image] | None
    spectra: dict[str, Spectrum] | None



import logging
import pathlib
from functools import partial

import numpy as np

import qtpy
from qtpy import QtGui, QtCore, QtWidgets


from ..utils import AbstractWidget
from ..appdata import AppData


logger = logging.getLogger(__name__)


class ControlBar(QtWidgets.QToolBar, AbstractWidget):
    reset_view_button_clicked = QtCore.Signal()
    reset_dock_state_button_clicked = QtCore.Signal()
    screenshot_button_clicked = QtCore.Signal(str)
    object_selected = QtCore.Signal(int)

    def __init__(self, rd: AppData, parent=None):
        super().__init__(parent=parent)
        self.setWindowTitle('Control Panel')

        self.rd = rd

        # create buttons for switching to the next or previous object
        self._pn_buttons = self.create_pn_buttons()

        # create the `starred` button
        self._star_button = QtWidgets.QAction(parent=self)
        self._star_button.setIcon(self.get_icon(self.get_star_icon_name()))
        self._star_button.setToolTip('Star the object')
        self._star_button.triggered.connect(self.star)

        # create the `screenshot` button
        self._screenshot_button = QtWidgets.QAction(parent=self)
        self._screenshot_button.setIcon(self.get_icon('screenshot.svg'))
        self._screenshot_button.setToolTip('Take a screenshot')
        self._screenshot_button.triggered.connect(self.screenshot)

        # create the `reset view` button
        self._reset_view_button = QtWidgets.QAction(parent=self)
        self._reset_view_button.setIcon(self.get_icon('reset-view.svg'))
        self._reset_view_button.setToolTip('Reset the view')
        self._reset_view_button.triggered.connect(self.reset_view_button_clicked.emit)

        # create a `dark mode` button
        # self._dark_mode = QtWidgets.QPushButton(parent=self)
        # self._dark_mode.setIcon(self.create_icon(get_icon_abs_path('dark-mode.svg')))
        # self._dark_mode.setToolTip('Turn on the dark theme')

        # create the `reset dock state` button
        self._reset_dock_state_button = QtWidgets.QAction(parent=self)
        self._reset_dock_state_button.setIcon(self.get_icon('reset-dock-state.svg'))
        self._reset_dock_state_button.setToolTip('Reset the dock state')
        self._reset_dock_state_button.triggered.connect(self.reset_dock_state_button_clicked.emit)

    def create_pn_buttons(self) -> dict[str, QtWidgets.QAction]:
        pn_buttons_params = {'previous': {'shortcut': QtGui.QKeySequence.MoveToPreviousChar, 'icon': 'arrow-backward'},
                             'next': {'shortcut': QtGui.QKeySequence.MoveToNextChar, 'icon': 'arrow-forward'},
                             'previous starred': {'icon': 'arrow-backward-starred'},
                             'next starred': {'icon': 'arrow-forward-starred'}
                             }

        pn_buttons = {}
        for pn_text, pn_properties in pn_buttons_params.items():
            button = QtWidgets.QAction('Go to the {} object'.format(pn_text), self)
            button.setIcon(self.get_icon(pn_properties['icon'] + '.svg'))

            button.triggered.connect(partial(self.previous_next_object, pn_text.split(' ')[0], 'starred' in pn_text))
            if pn_properties.get('shortcut'):
                button.setShortcut(pn_properties['shortcut'])

            pn_buttons[pn_text] = button

        return pn_buttons

    def init_ui(self):
        self.addAction(self._pn_buttons['previous starred'])
        self.addAction(self._pn_buttons['previous'])
        self.addAction(self._pn_buttons['next'])
        self.addAction(self._pn_buttons['next starred'])

        self.addSeparator()

        self.addAction(self._screenshot_button)
        self.addAction(self._star_button)

        self.addSeparator()

        self.addAction(self._reset_view_button)
        self.addAction(self._reset_dock_state_button)

    def load_object(self):
        self._star_button.setIcon(self.get_icon(self.get_star_icon_name(self.rd.df.at[self.rd.id, 'starred'])))

        self._pn_buttons['previous starred'].setEnabled(np.sum(self.rd.df['starred']) > 0)
        self._pn_buttons['next starred'].setEnabled(np.sum(self.rd.df['starred']) > 0)

    def previous_next_object(self, command: str, starred: bool):
        j_upd = self.update_index(self.rd.j, self.rd.n_objects, command)

        if starred:
            if np.sum(self.rd.df['starred']) > 0:
                while not self.rd.df.iat[j_upd, self.rd.df.columns.get_loc('starred')]:
                    j_upd = self.update_index(j_upd, self.rd.n_objects, command)
            else:
                return

        self.object_selected.emit(j_upd)

    @staticmethod
    def update_index(current_index, n_objects, command: str):
        j_upd = current_index

        if command == 'next':
            j_upd += 1
        elif command == 'previous':
            j_upd -= 1

        j_upd = j_upd % n_objects

        return j_upd

    def get_icon_abs_path(self, icon_name: str) -> pathlib.Path:
        icon_root_dir = pathlib.Path(__file__).parent.parent / 'data' / 'icons'
        icon_path = icon_root_dir / self.rd.config.appearance.theme / icon_name
        if not icon_path.exists():
            return icon_root_dir / 'light' / icon_name
        else:
            return icon_path

    @staticmethod
    def get_star_icon_name(starred=False):
        icon_name = 'star.svg' if starred else 'star-empty.svg'
        return icon_name

    def get_icon(self, icon_name):
        return QtGui.QIcon(str(self.get_icon_abs_path(icon_name)))

    def star(self):
        starred = not self.rd.df.at[self.rd.id, 'starred']

        self.rd.df.at[self.rd.id, 'starred'] = starred
        self._star_button.setIcon(self.get_icon(self.get_star_icon_name(starred)))

        self._pn_buttons['previous starred'].setEnabled(np.sum(self.rd.df['starred']) > 0)
        self._pn_buttons['next starred'].setEnabled(np.sum(self.rd.df['starred']) > 0)

    def screenshot(self):
        default_filename = '{}_ID{}.png'.format(self.rd.output_path.stem.replace(' ', '_'), self.rd.id)
        path, extension = qtpy.compat.getsavefilename(self, caption='Save/Save As',
                                                      basedir=str(pathlib.Path().resolve() / default_filename),
                                                      filters='Images (*.png)')

        if path:
            self.screenshot_button_clicked.emit(path)



import logging
import abc
import pathlib
from dataclasses import asdict

import numpy as np
from astropy.table import Table
from astropy.io.fits.header import Header

from qtpy import QtWidgets

from .LazyViewerElement import LazyViewerElement
from ..utils import SmartSliderWithEditor
from ..appdata import AppData
from ..config import docks
from ..io.viewer_data import get_filename, load


logger = logging.getLogger(__name__)


class ViewerElement(LazyViewerElement, abc.ABC):
    def __init__(self, cfg: docks.ViewerElement, **kwargs):
        self.lazy_widgets: list[LazyViewerElement] = []
        self.sliders: list[SmartSliderWithEditor] = []

        super().__init__(cfg=cfg, **kwargs)

        self.cfg = cfg

        self.filename: pathlib.Path | None = None
        self.data = None
        self.meta: dict | Header | None = None
        self.allowed_data_types: tuple | None = None

        # create a smoothing slider
        self.smoothing_slider = self.create_smoothing_slider(**asdict(self.cfg.smoothing_slider))
        self.sliders.append(self.smoothing_slider)

    def create_smoothing_slider(self, **kwargs):
        smoothing_slider = SmartSliderWithEditor(parameter='sigma', action='smooth the data', parent=self, **kwargs)
        smoothing_slider.value_changed[float].connect(self.smooth)
        smoothing_slider.setToolTip('Slide to smooth the data')
        return smoothing_slider

    def init_ui(self):
        sub_layout = QtWidgets.QHBoxLayout()

        # add vertical sliders
        for s in self.sliders:
            if not s.text_editor:
                sub_layout.addWidget(s)
        sub_layout.addWidget(self.graphics_view)  # add the central widget
        self.layout().addLayout(sub_layout, 1, 1, 1, 1)

        sub_layout = QtWidgets.QVBoxLayout()

        # add horizontal sliders
        for s in self.sliders:
            if s.text_editor:
                sub_layout.addWidget(s)
        self.layout().addLayout(sub_layout, 2, 1, 1, 1)

        # init the UI of lazy widgets
        for w in self.lazy_widgets:
            w.init_ui()

        # init the UI of sliders
        for s in self.sliders:
            s.init_ui()

    def load_object(self, rd: AppData):
        # clear the widget content
        if self.data is not None:
            self.clear_content()

            for s in self.sliders:
                s.clear()

        # load catalogue values to the sliders
        for s in self.sliders:
            if s.cat_name is not None:
                s.update_default_value(rd.cat, rd.id, rd.config.catalogue.translate)

        # load data to the widget
        self._load_data(rd=rd)

        # display the data
        if self.data is not None:
            self.setEnabled(True)
            self.display()

            for s in self.sliders:
                s.update_from_slider()

            self.reset_view()
        else:
            self.setEnabled(False)

    def _load_data(self, rd: AppData):
        self.filename = get_filename(rd.config.data.dir, self.cfg.filename_keyword, rd.id)

        if self.filename is None:
            logger.error('{} not found (object ID: {})'.format(self.title, rd.id))
            self.data, self.meta = None, None
            return

        loader_config = {} if self.cfg.data_loader_params is None else self.cfg.data_loader_params

        self.data, self.meta = load(self.cfg.data_loader, self.filename, self.title, **loader_config)
        if self.data is None:
            return

        if self.allowed_data_types is not None:
            if not any(isinstance(self.data, t) for t in self.allowed_data_types):
                logger.error(f'Invalid input data type: {type(self.data)} (widget: {self.title}). '
                             'Try to use a different data loader')
                self.data, self.meta = None, None
                return

    def setEnabled(self, a0: bool = True):
        super().setEnabled(a0)
        for w in self.lazy_widgets:
            w.setEnabled(a0)
        for s in self.sliders:
            s.setEnabled(a0)

    @abc.abstractmethod
    def display(self):
        pass

    @abc.abstractmethod
    def reset_view(self):
        pass

    @abc.abstractmethod
    def clear_content(self):
        pass

    @abc.abstractmethod
    def smooth(self, sigma: float):
        pass
